/*
 * Matthew Avery Coder 2012 - 2013.
 * Copyright (c) 2013, Michael Hanselmann.
 * Distributed under the Boost Software License, Version 1.0.
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer, must
 * be included in all copies of the Software, in whole or in part, and all
 * derivative works of the Software, unless such copies or derivative works are
 * solely in the form of machine-executable object code generated by a source
 * language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <cassert>
#include <limits>
#include <stdexcept>
#include <cctype>

#include "base64.hpp"

namespace
{
const char kBase64Table[] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
}

namespace QrBackup
{
size_t CalcBase64PaddingLength(const size_t length)
{
  return (length % 3u) ? (3u - (length % 3u)) : 0u;
}

size_t CalcBase64Length(const size_t length)
{
  const size_t padding_size = CalcBase64PaddingLength(length);

  return 4u * ((length + padding_size) / 3u);
}

size_t CalcBase64Capacity(const size_t length)
{
  return (length / 4u) * 3u;
}

std::string EncodeBase64(const unsigned char *const data, const size_t length)
{
  const size_t code_size = CalcBase64Length(length);

  if (code_size > std::numeric_limits<std::string::size_type>::max()) {
    throw std::length_error("Converting too large a string to base64");
  }

  std::string retval;

  /* Pre-allocate space including padding */
  retval.assign(code_size, '=');

  const unsigned char *data_end = data + length;
  size_t outpos = 0;
  int bits_collected = 0;
  unsigned int accumulator = 0;

  for (const unsigned char *i = data; i != data_end; ++i) {
    accumulator = (accumulator << 8) | (*i & 0xffu);
    bits_collected += 8;
    while (bits_collected >= 6) {
      bits_collected -= 6;
      retval[outpos++] = kBase64Table[(accumulator >> bits_collected) & 0x3fu];
    }
  }

  assert(outpos <= retval.size());

  if (bits_collected > 0) { // Are any trailing bits left over?
    assert(bits_collected < 6);
    accumulator <<= 6 - bits_collected;
    retval[outpos++] = kBase64Table[accumulator & 0x3fu];
  }

  assert(outpos == (retval.size() - CalcBase64PaddingLength(length)));
  assert(outpos <= retval.size());

  return retval;
}
}

/* vim: set sw=2 sts=2 et : */
