/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <algorithm>
#include <array>
#include <cassert>
#include <sstream>

#include <qrencode.h>

#include "capacity.hpp"

namespace
{
const std::array<std::array<size_t, QRSPEC_VERSION_MAX + 1>, 4> kCapacity{{
    {0, 17, 32, 53, 78, 106, 134, 154, 192, 230, 271, 321, 367, 425, 458, 520,
     586, 644, 718, 792, 858, 929, 1003, 1091, 1171, 1273, 1367, 1465, 1528,
     1628, 1732, 1840, 1952, 2068, 2188, 2303, 2431, 2563, 2699, 2809, 2953},
    {0, 14, 26, 42, 62, 84, 106, 122, 152, 180, 213, 251, 287, 331, 362, 412,
     450, 504, 560, 624, 666, 711, 779, 857, 911, 997, 1059, 1125, 1190, 1264,
     1370, 1452, 1538, 1628, 1722, 1809, 1911, 1989, 2099, 2213, 2331},
    {0, 11, 20, 32, 46, 60, 74, 86, 108, 130, 151, 177, 203, 241, 258, 292, 322,
     364, 394, 442, 482, 509, 565, 611, 661, 715, 751, 805, 868, 908, 982, 1030,
     1112, 1168, 1228, 1283, 1351, 1423, 1499, 1579, 1663},
    {0, 7, 14, 24, 34, 44, 58, 64, 84, 98, 119, 137, 155, 177, 194, 220, 250,
     280, 310, 338, 382, 403, 439, 461, 511, 535, 593, 625, 658, 698, 742, 790,
     842, 898, 958, 983, 1051, 1093, 1139, 1219, 1273},
}};
}

namespace QrBackup
{
/* Chooses the smallest QR code version which can fit the given amout of data.
 * If the data is too large for a single QR code, the largest version is
 * returned.
 */
QrVersion ChooseQrVersion(const size_t size, const int eclevel)
{
  assert(eclevel >= 0 && eclevel < int{kCapacity.size()});

  const auto &values = kCapacity[eclevel];

  const size_t maxcapacity = values[values.size() - 1];

  if (size >= maxcapacity) {
    return {QRSPEC_VERSION_MAX, maxcapacity};
  }

  const auto it = std::upper_bound(values.cbegin(), values.cend(), size);

  if (it == values.cend()) {
    std::ostringstream buf{"Unable to find upper bound for size"};
    buf << size << " and level " << eclevel;

    throw std::range_error{buf.str()};
  }

  return {std::distance(values.cbegin(), it), *it};
}
}

/* vim: set sw=2 sts=2 et : */
