/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <memory>
#include <sstream>
#include <iomanip>
#include <iterator>
#include <stdexcept>

#include <openssl/evp.h>

#include "digest.hpp"

namespace
{
struct Deleter
{
  void operator()(EVP_MD_CTX *ctx)
  {
    EVP_MD_CTX_destroy(ctx);
  }
};

std::string CalcDigestInner(const void *const message, const size_t length,
                            const EVP_MD *const md)
{
  using QrBackup::Hexify;

  unsigned char md_value[EVP_MAX_MD_SIZE];
  unsigned int md_len;

  std::unique_ptr<EVP_MD_CTX, Deleter> ctx;

  ctx.reset(EVP_MD_CTX_create());

  if (EVP_DigestInit_ex(ctx.get(), md, nullptr) != 1) {
  error:
    throw std::runtime_error{"Digest calculation failed"};
  }

  if (EVP_DigestUpdate(ctx.get(), message, length) != 1) {
    goto error;
  }

  if (EVP_DigestFinal_ex(ctx.get(), md_value, &md_len) != 1) {
    goto error;
  }

  return Hexify(md_value, md_len);
}
}

namespace QrBackup
{
HashAlgorithm::HashAlgorithm() : _nid{NID_undef}
{
}

HashAlgorithm::HashAlgorithm(const int nid) : _nid{nid}
{
}

int HashAlgorithm::nid() const
{
  return _nid;
}

std::string HashAlgorithm::name() const
{
  return OBJ_nid2sn(_nid);
}

bool operator==(const HashAlgorithm &lhs, const HashAlgorithm &rhs)
{
  return lhs.nid() == rhs.nid();
}

bool operator<(const HashAlgorithm &lhs, const HashAlgorithm &rhs)
{
  return lhs.name() < rhs.name();
}

std::map<HashAlgorithm, std::string> CalcDigest(const void *message,
                                                const size_t length)
{
  static const EVP_MD *const kAlgorithms[]{
#ifndef OPENSSL_NO_MD5
      EVP_md5(),
#endif
#ifndef OPENSSL_NO_SHA
      EVP_sha1(),
#endif
#ifndef OPENSSL_NO_SHA256
      EVP_sha256(),
#endif
#ifndef OPENSSL_NO_WHIRLPOOL
      EVP_whirlpool(),
#endif
#ifndef OPENSSL_NO_RIPEMD
      EVP_ripemd160(),
#endif
  };

  std::map<HashAlgorithm, std::string> result;

  for (const auto md : kAlgorithms) {
    result[EVP_MD_nid(md)] = CalcDigestInner(message, length, md);
  }

  return result;
}

/* Convert arbitrary data into hex string.
 */
std::string Hexify(const unsigned char *data, const unsigned int len)
{
  std::ostringstream buf;

  buf << std::hex << std::setfill('0') << std::nouppercase;

  for (unsigned int i = 0; i < len; ++i) {
    /* setw() must be repeated for every byte */
    buf << std::setw(2) << (int)data[i];
  }

  return buf.str();
}
}

/* vim: set sw=2 sts=2 et : */
