/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <cassert>
#include <memory>
#include <iterator>

#include "file.hpp"
#include "capacity.hpp"
#include "base64.hpp"

namespace QrBackup
{
namespace
{
std::vector<std::shared_ptr<FilePart>> MakeParts(
    const QRecLevel eclevel, const std::vector<unsigned char> &content)
{
  std::vector<std::shared_ptr<FilePart>> result;

  /* Split and encode content */
  for (auto it = content.begin(); it != content.end();) {
    const size_t remaining = std::distance(it, content.end());
    int version;
    size_t version_size;

    /* Select QR code version */
    std::tie(version, version_size) =
        ChooseQrVersion(CalcBase64Length(remaining), eclevel);

    const size_t partlength =
        std::min(CalcBase64Capacity(version_size), remaining);

    /* Encode in Base64 */
    const std::string part_base64(EncodeBase64(&*it, partlength));

    assert(part_base64.size() == CalcBase64Length(partlength));

    /* Generate QR code */
    auto qrcode = std::make_unique<QrCode>(part_base64, eclevel, version);

    /* Hash part */
    const auto digest = CalcDigest(&*it, partlength);

    /* Append to list of parts */
    auto part = std::make_shared<FilePart>(std::distance(content.begin(), it),
                                           partlength, part_base64.length(),
                                           digest, std::move(qrcode));

    result.emplace_back(part);

    std::advance(it, partlength);
  }

  return result;
}
}

FileContext::FileContext(const QRecLevel eclevel,
                         const std::string &description,
                         const std::vector<unsigned char> &content)
    : _description(description), _length(content.size())
{
  _content_digest = CalcDigest(&content[0], content.size());
  _parts = MakeParts(eclevel, content);
}
}

/* vim: set sw=2 sts=2 et : */
