/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <vector>
#include <memory>

#include "qrcode.hpp"
#include "digest.hpp"

namespace QrBackup
{
class FilePart final
{
public:
  FilePart(const size_t offset, const size_t length,
           const size_t encoded_length,
           const std::map<HashAlgorithm, std::string> &digest,
           std::unique_ptr<QrCode> qrcode)
      : offset{offset}, length{length}, encoded_length{encoded_length},
        digest{digest}, qrcode{std::move(qrcode)}
  {
  }
  FilePart(const FilePart &) = delete;

  FilePart &operator=(const FilePart &) = delete;

  size_t offset;
  size_t length;
  size_t encoded_length;
  std::map<HashAlgorithm, std::string> digest;
  std::unique_ptr<QrCode> qrcode;
};

class FileContext final
{
public:
  FileContext() = delete;
  FileContext(const QRecLevel eclevel, const std::string &description,
              const std::vector<unsigned char> &content);
  FileContext(const FileContext &) = delete;

  FileContext &operator=(const FileContext &) = delete;

  std::string description() const
  {
    return _description;
  }

  size_t length() const
  {
    return _length;
  }

  std::map<HashAlgorithm, std::string> content_digest() const
  {
    return _content_digest;
  }

  std::vector<std::shared_ptr<FilePart>> parts() const
  {
    return _parts;
  }

private:
  std::string _description;
  size_t _length;

  std::map<HashAlgorithm, std::string> _content_digest;
  std::vector<std::shared_ptr<FilePart>> _parts;
};
}

/* vim: set sw=2 sts=2 et : */
