/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <cassert>

#include <cstdlib>
#include <cstring>
#include <cerrno>

#include <unistd.h>

#include <pango/pangocairo.h>
#include <openssl/opensslv.h>
#include <openssl/crypto.h>

#include "options.hpp"
#include "qrcode.hpp"
#include "utils.hpp"

namespace
{
const ::QRecLevel kDefaultEcLevel{QR_ECLEVEL_M};
}

namespace QrBackup
{
ProgramOptions::ProgramOptions() : eclevel{kDefaultEcLevel}
{
}

void Usage(std::ostream &output, const std::string &progname)
{
  output << "Usage: " << progname
         << " [options] { <input> | - } { <output> | - }\n" << '\n'
         << "Options:\n"
         << "  -h  Display this usage description and exit\n"
         << "  -V  Output version information and exit\n"
         << "  -d  Descriptive text\n"
         << "  -l  Error correction level (one of ";

  {
    bool first{true};
    for (const auto &name : GetEcLevelNames()) {
      if (!first) {
        output << ", ";
      }
      output << name;
      first = false;
    }
  }

  {
    const auto def = GetEcLevelName(kDefaultEcLevel);
    assert(def.first);
    output << "; defaults to " << def.second << ")\n";
  }

  output << '\n' << "Using libqrencode " << ::QRcode_APIVersionString() << ", "
         << "Cairo 2D " << ::cairo_version_string() << ", "
         << "Pango " << ::pango_version_string() << " and "
         << "OpenSSL " << ::SSLeay_version(SSLEAY_VERSION) << ".\n";

  output.flush();
}

void VersionInfo(std::ostream &output, const std::string &progname)
{
  output << progname << ' ' << PACKAGE_VERSION << '\n';
  output.flush();
}

int ParseArgs(int argc, char *const *argv, ProgramOptions &options,
              std::ostream &normal_output, std::ostream &error_output)
{
  const std::string progname =
      ((argc > 0) ? GetProgramName(argv[0]) : "(unknown)");
  int opt;

  /* Required by GNU libc's getopt(3) when called more than once during a
   * program's lifetime (such as unit tests).
   */
  optind = 0;

  while ((opt = ::getopt(argc, argv, ":hVqd:l:")) != -1) {
    switch (opt) {
    case 'd':
      options.description = optarg;
      break;

    case 'l': {
      const auto tmp = LookupEcLevel(optarg);
      if (!tmp.first) {
        Usage(error_output, progname);
        return EXIT_FAILURE;
      }
      options.eclevel = tmp.second;
      break;
    }

    case 'h':
      Usage(normal_output, progname);
      return EXIT_SUCCESS;

    case 'V':
      VersionInfo(normal_output, progname);
      return EXIT_SUCCESS;

    default:
      Usage(error_output, progname);
      return EXIT_FAILURE;
    }
  }

  /* Shift away options and values */
  argc -= optind;
  argv += optind;

  if (argc != 2) {
    Usage(error_output, progname);
    return EXIT_FAILURE;
  }

  options.input_filename = argv[0];
  options.output_filename = argv[1];

  return -1;
}
}

/* vim: set sw=2 sts=2 et : */
