/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sstream>
#include <algorithm>
#include <stdexcept>
#include <iostream>

#include <pango/pangocairo.h>
#include <cairo-pdf.h>

#include "pdf.hpp"
#include "conversion.hpp"
#include "utils.hpp"

namespace
{
using TableContent =
    std::vector<std::pair<const std::string, const std::string>>;

const double kFontSize = 8.0;
const double kParagraphSpacing = QrBackup::CmToPoint(.5);

struct Border
{
  double top, right, bottom, left;
};

struct Paper
{
  double width;
  double height;
  Border margin;
};

struct Deleter
{
  void operator()(PangoLayout *ptr)
  {
    g_object_unref(ptr);
  }

  void operator()(PangoFontDescription *ptr)
  {
    ::pango_font_description_free(ptr);
  }

  void operator()(cairo_t *ptr)
  {
    ::cairo_destroy(ptr);
  }

  void operator()(cairo_surface_t *ptr)
  {
    ::cairo_surface_destroy(ptr);
  }
};

std::string FormatError(const std::string &prefix,
                        const ::cairo_status_t status)
{
  std::ostringstream buf{prefix};
  buf << ": " << ::cairo_status_to_string(status);
  return buf.str();
}

class CairoError final : public std::runtime_error
{
public:
  CairoError(const std::string &prefix, const ::cairo_status_t status)
      : std::runtime_error{FormatError(prefix, status)}
  {
  }
};

void CheckStatus(const ::cairo_status_t status, const std::string &msg)
{
  if (status != CAIRO_STATUS_SUCCESS) {
    throw CairoError(msg, status);
  }
}

void CheckStatus(::cairo_t *const cr, const std::string &msg)
{
  CheckStatus(::cairo_status(cr), msg);
}

void CheckStatus(::cairo_surface_t *const surface, const std::string &msg)
{
  CheckStatus(::cairo_surface_status(surface), msg);
}

double GetLayoutHeight(::PangoLayout *const layout)
{
  int height;

  ::pango_layout_get_size(layout, NULL, &height);

  return ::pango_units_to_double(height);
}

double WriteLine(::cairo_t *const cr, ::PangoLayout *const layout,
                 const double left, const double top, const double width,
                 const std::string &text)
{
  ::cairo_move_to(cr, left, top);
  ::pango_cairo_update_layout(cr, layout);

  ::pango_layout_set_width(layout, ::pango_units_from_double(width));
  ::pango_layout_set_alignment(layout, PANGO_ALIGN_LEFT);
  ::pango_layout_set_wrap(layout, PANGO_WRAP_WORD_CHAR);
  ::pango_layout_set_text(layout, text.c_str(), -1);
  ::pango_cairo_show_layout(cr, layout);

  return top + GetLayoutHeight(layout);
}

double WriteQrCode(::cairo_t *const cr, ::PangoLayout *const layout,
                   const double left, const double top, const double width,
                   const double height, const QrBackup::QrCode &qrcode)
{
  const unsigned int qrmodcount = qrcode.width();

  const double qrwidth = std::min(width, height);
  const double qrmodwidth = qrwidth / qrmodcount;

  const double qrleft = left + ((width - qrwidth) / 2);

  const auto data = qrcode.data();

  ::cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
  ::cairo_set_line_width(cr, 0);

  for (unsigned int y = 0; y < qrmodcount; ++y) {
    for (unsigned int x = 0; x < qrmodcount; ++x) {
      if (data[(y * qrmodcount) + x]) {
        const int begin = x;

        /* Merge columns to reduce drawing operations (and file size) */
        for (; x < qrmodcount && data[(y * qrmodcount) + x]; ++x) {
        }

        ::cairo_rectangle(cr, qrleft + (begin * qrmodwidth),
                          top + (y * qrmodwidth), (x - begin) * qrmodwidth,
                          qrmodwidth);
      }
    }
  }

  ::cairo_fill(cr);

  return top + height;
}
}

namespace QrBackup
{
static ::cairo_status_t PdfWriteCb(void *closure, const unsigned char *data,
                                   unsigned int length);

namespace
{
const Paper kDefaultPaper{
    /* Keep content small enough to fit on A4 and Letter paper */
    std::min(CmToPoint(21), InchToPoint(8.5)),
    std::min(CmToPoint(29.7), InchToPoint(11)),
    {CmToPoint(1), CmToPoint(1), CmToPoint(1), CmToPoint(1)},
};
}

void CheckPangoVersion()
{
  const char *msg = pango_version_check(
      PANGO_VERSION_MAJOR, PANGO_VERSION_MINOR, PANGO_VERSION_MICRO);

  if (msg != NULL) {
    std::ostringstream buf{"Pango version check failed: "};
    buf << msg;
    throw std::runtime_error{buf.str()};
  }
}

static double WriteTable(::cairo_t *const cr, ::PangoLayout *const layout,
                         const double left, double top, const double width,
                         const TableContent &rows)
{
  int longest_label = 0;

  /* Find longest label */
  for (const auto &i : rows) {
    const std::string label = i.first + " ";
    int label_width;

    ::pango_layout_set_width(layout, ::pango_units_from_double(width));
    ::pango_layout_set_text(layout, label.c_str(), -1);

    ::pango_layout_get_size(layout, &label_width, NULL);

    if (label_width > longest_label) {
      longest_label = label_width;
    }
  }

  /* Write rows */
  for (const auto &i : rows) {
    const std::string &label = i.first;
    const std::string &value = i.second;

    /* Label */
    ::cairo_move_to(cr, left, top);
    ::pango_cairo_update_layout(cr, layout);

    ::pango_layout_set_width(layout, -1);
    ::pango_layout_set_text(layout, label.c_str(), -1);
    ::pango_cairo_show_layout(cr, layout);

    const double label_height = GetLayoutHeight(layout);

    /* Value */
    ::cairo_move_to(cr, left + ::pango_units_to_double(longest_label), top);
    ::pango_cairo_update_layout(cr, layout);

    ::pango_layout_set_width(layout,
                             ::pango_units_from_double(width) - longest_label);
    ::pango_layout_set_wrap(layout, PANGO_WRAP_WORD_CHAR);
    ::pango_layout_set_text(layout, value.c_str(), -1);
    ::pango_cairo_show_layout(cr, layout);

    top += std::max(::pango_units_to_double(label_height),
                    GetLayoutHeight(layout));
  }

  return top;
}

static double WriteLabel(::cairo_t *const cr, ::PangoLayout *const layout,
                         const double left, const double top,
                         const double width, const std::string &label,
                         const std::string &value)
{
  TableContent rows;

  rows.push_back(std::make_pair(label + ":", value));

  return WriteTable(cr, layout, left, top, width, rows);
}

static double WriteHashes(::cairo_t *const cr, ::PangoLayout *const layout,
                          const double left, const double top,
                          const double width,
                          const std::map<HashAlgorithm, std::string> &hashes)
{
  TableContent rows;

  /* Make table from hashes */
  for (const auto &i : hashes) {
    rows.push_back(std::make_pair(i.first.name() + ":", i.second));
  }

  return WriteTable(cr, layout, left, top, width, rows);
}

class PdfPage final
{
public:
  PdfPage(::cairo_t *const cr, const Paper &paper)
      : _cr{cr}, _paper{paper}, _layout{::pango_cairo_create_layout(cr)}
  {
    /* Font */
    const std::unique_ptr<PangoFontDescription, Deleter> desc{
        ::pango_font_description_from_string("monospace")};

    ::pango_font_description_set_size(desc.get(), kFontSize * PANGO_SCALE);

    ::pango_layout_set_font_description(_layout.get(), desc.get());
  }

  PdfPage(const PdfPage &) = delete;
  PdfPage &operator=(const PdfPage &) = delete;

  void operator()(const unsigned int page_count, const unsigned int page_number,
                  const std::string &timestamp, const std::string &description,
                  const size_t content_length,
                  const std::map<HashAlgorithm, std::string> &content_digest,
                  const size_t part_offset, const size_t part_length,
                  const size_t part_encoded_length,
                  const std::map<HashAlgorithm, std::string> &part_digest,
                  const QrCode &qrcode)
  {
    const double available_width =
        _paper.width - _paper.margin.left - _paper.margin.right;
    double top = _paper.margin.top;

    /* White background */
    ::cairo_reset_clip(_cr);
    ::cairo_set_source_rgb(_cr, 1.0, 1.0, 1.0);
    ::cairo_paint(_cr);

    /* Black colour */
    ::cairo_set_source_rgb(_cr, 0, 0, 0);

    /* Timestamp */
    {
      ::cairo_move_to(_cr, _paper.margin.left, top);
      ::pango_cairo_update_layout(_cr, _layout.get());

      ::pango_layout_set_width(_layout.get(),
                               ::pango_units_from_double(available_width));
      ::pango_layout_set_alignment(_layout.get(), PANGO_ALIGN_RIGHT);
      ::pango_layout_set_text(_layout.get(), timestamp.c_str(), -1);
      ::pango_cairo_show_layout(_cr, _layout.get());
    }

    /* File information */
    top = WriteLine(_cr, _layout.get(), _paper.margin.left, top,
                    available_width, "File information:");

    /* Description */
    if (!description.empty()) {
      top = WriteLabel(_cr, _layout.get(), _paper.margin.left, top,
                       available_width, "Description", description);
    }

    /* Length */
    {
      std::ostringstream buf;

      buf << content_length << " bytes";

      top = WriteLabel(_cr, _layout.get(), _paper.margin.left, top,
                       available_width, "Length", buf.str());
    }

    /* File hash sums */
    top = WriteHashes(_cr, _layout.get(), _paper.margin.left, top,
                      available_width, content_digest);

    top += kParagraphSpacing;

    if (page_count > 1) {
      /* Part information */
      {
        std::ostringstream buf;

        buf << "Part " << page_number << " of " << page_count << ":";

        top = WriteLine(_cr, _layout.get(), _paper.margin.left, top,
                        available_width, buf.str());
      }

      {
        std::ostringstream buf;

        buf << "Range: " << part_offset << "-"
            << (part_offset + part_length - 1) << ", "
            << "length " << part_length << " bytes";

        top = WriteLine(_cr, _layout.get(), _paper.margin.left, top,
                        available_width, buf.str());
      }

      top = WriteHashes(_cr, _layout.get(), _paper.margin.left, top,
                        available_width, part_digest);

      top += kParagraphSpacing;
    }

    /* QR code details */
    {
      std::ostringstream buf;

      buf << "QR code version " << qrcode.version() << ", " << qrcode.width()
          << "\xC3\x97" << qrcode.width() << " modules, "
          << "error correction level " << qrcode.eclevel() << ". "
          << "Data encoded using Base64 without linebreaks ("
          << part_encoded_length << " bytes after encoding).";

      top = WriteLine(_cr, _layout.get(), _paper.margin.left, top,
                      available_width, buf.str());
    }

    top += 2 * kParagraphSpacing;

    /* QR code */
    top = WriteQrCode(_cr, _layout.get(), _paper.margin.left, top,
                      available_width,
                      (_paper.height - _paper.margin.bottom - top), qrcode);

    /* Output page */
    ::cairo_show_page(_cr);
  }

private:
  ::cairo_t *const _cr;
  const Paper &_paper;
  const std::unique_ptr<::PangoLayout, Deleter> _layout;
};

struct PdfOutputImpl final
{
  OutputWriteFn &write_fn;
  const Paper &paper;
  const std::string timestamp;
  const std::string &description;

  std::unique_ptr<::cairo_surface_t, Deleter> surface;
  std::unique_ptr<::cairo_t, Deleter> cr;
  unsigned int page_number;

  PdfOutputImpl(OutputWriteFn &write_fn, const Paper &paper,
                const std::string &timestamp, const std::string &description)
      : write_fn{write_fn}, paper{paper}, timestamp{timestamp},
        description{description},
        surface{::cairo_pdf_surface_create_for_stream(
            PdfWriteCb, this, paper.width, paper.height)},
        cr{::cairo_create(surface.get())}, page_number{1}
  {
    CheckStatus(surface.get(), "Failed to create Cairo PDF surface");
    CheckStatus(cr.get(), "Failed to create Cairo drawing context");
  }
};

static ::cairo_status_t PdfWriteCb(void *closure, const unsigned char *data,
                                   unsigned int length)
{
  const PdfOutputImpl *const pimpl = static_cast<PdfOutputImpl *>(closure);

  try {
    if (!pimpl->write_fn(data, length)) {
      return CAIRO_STATUS_WRITE_ERROR;
    }
  } catch (const std::exception &exc) {
    std::cerr << exc.what() << std::endl;
    return CAIRO_STATUS_WRITE_ERROR;
  }

  return CAIRO_STATUS_SUCCESS;
}

PdfOutput::PdfOutput(OutputWriteFn &write_fn, const FileContext &fctx)
    : OutputBase{fctx},
      pimpl{std::make_unique<PdfOutputImpl>(
          write_fn, kDefaultPaper, CurrentTimeWithTz(), fctx.description())}
{
}

PdfOutput::~PdfOutput()
{
}

void PdfOutput::WritePart(const FilePart &part)
{
  const Paper &paper = pimpl->paper;
  ::cairo_t *const cr = pimpl->cr.get();

  PdfPage(cr, paper)(fctx.parts().size(), pimpl->page_number, pimpl->timestamp,
                     fctx.description(), fctx.length(), fctx.content_digest(),
                     part.offset, part.length, part.encoded_length, part.digest,
                     *part.qrcode.get());

  ::cairo_surface_flush(pimpl->surface.get());

  /* Check error status */
  {
    std::ostringstream buf{"Drawing page "};
    buf << pimpl->page_number << " failed";

    CheckStatus(cr, buf.str());
    CheckStatus(pimpl->surface.get(), buf.str());
  }

  ++pimpl->page_number;
}

void PdfOutput::Finalize()
{
  ::cairo_surface_finish(pimpl->surface.get());

  CheckStatus(pimpl->surface.get(), "Finalizing surface failed");

  pimpl->surface.reset();
  pimpl->cr.reset();
}
}

/* vim: set sw=2 sts=2 et : */
