/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <vector>
#include <system_error>
#include <sstream>
#include <iterator>
#include <algorithm>

#include <cstdlib>
/* For P_tmpdir */
#include <cstdio>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include "posixfile.hpp"

namespace
{
/* Candidate directories for creating temporary files
 */
const std::string kTempDirs[] = {
#if defined(P_tmpdir)
    P_tmpdir,
#endif
    "/tmp",
};

int WrapOpen(const std::string &path, const int flags, const int mode)
{
  const int fd = ::open(path.c_str(), flags, mode);

  if (fd < 0) {
    const std::error_code errcode{errno, std::system_category()};

    std::ostringstream buf{"open("};
    buf << path << ")";

    throw std::system_error{errcode, buf.str()};
  }

  return fd;
}

void MaybeClose(const int fd)
{
  if (fd >= 0) {
    ::close(fd);
  }
}
}

namespace QrBackup
{
PosixFile::PosixFile(const std::string &path, const int flags, const int mode)
    : FileDescriptor{WrapOpen(path, flags, mode)}
{
}

PosixFile::~PosixFile()
{
  MaybeClose(fd);
}

/* Build a list of templates to be used for mkstemp(3).
 */
std::vector<std::string> GetTempNameTemplates()
{
  std::vector<std::string> result;

  for (const auto &i : kTempDirs) {
    result.emplace_back(i + "/qrbackup.XXXXXXXX");
  }

  return result;
}

/* Attempt to create a temporary file using mkstemp(3). If no file can be
 * created, an exception is thrown. Otherwise the file descriptor of the
 * unlinked file is returned.
 */
int WrapMkstemp(const std::vector<std::string> &templates)
{
  if (templates.empty()) {
    throw std::runtime_error{"List of templates can't be empty"};
  }

  for (auto it = templates.begin();; ++it) {
    std::vector<char> path{std::begin(*it), std::end(*it)};
    path.emplace_back('\0');

    const int fd = ::mkstemp(&path[0]);

    if (fd >= 0) {
      if (::unlink(&path[0]) < 0) {
        const std::error_code errcode{errno, std::system_category()};
        std::ostringstream buf{"unlink("};
        buf << &path[0] << ")";
        throw std::system_error{errcode, buf.str()};
      }

      return fd;
    }

    if (std::distance(it, templates.end()) == 1) {
      /* Last template */
      throw std::system_error{errno, std::system_category(), "mkstemp"};
    }
  }

  throw std::logic_error{"Should never reach this line"};
}

TemporaryFile::TemporaryFile()
    : FileDescriptor{WrapMkstemp(GetTempNameTemplates())}
{
}

TemporaryFile::~TemporaryFile()
{
  MaybeClose(fd);
}
}

/* vim: set sw=2 sts=2 et : */
