/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <ostream>
#include <system_error>
#include <istream>
#include <iostream>
#include <memory>
#include <fstream>
#include <algorithm>
#include <vector>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#include "options.hpp"
#include "utils.hpp"
#include "posixfile.hpp"
#include "pdf.hpp"
#include "file.hpp"

namespace
{
const std::string kStdioFilename{"-"};
}

namespace QrBackup
{
class FdWriter final : public OutputWriteFn
{
public:
  FdWriter(const int fd) : _fd{fd}
  {
  }

  result_type operator()(first_argument_type &data,
                         second_argument_type &length)
  {
    ssize_t ret = ::write(_fd, data, length);

    if (ret < 0) {
      throw std::system_error{errno, std::system_category(), "write"};
    }

    return static_cast<size_t>(ret) == length;
  }

private:
  const int _fd;
};

static void Process(const ProgramOptions &options, const int input_fd,
                    const int output_fd)
{
  std::vector<unsigned char> content;

  /* Read whole file */
  ReadFile(input_fd, content);

  const FileContext fctx{options.eclevel, options.description, content};

  /* Write output */
  FdWriter writer{output_fd};

  PdfOutput output{writer, fctx};

  output.Generate();
}

int Main(int argc, char *argv[])
{
  /* Disable standard C++ streams as POSIX API will be used for I/O */
  std::cin.rdbuf(NULL);
  std::cout.rdbuf(NULL);
  std::wcin.rdbuf(NULL);
  std::wcout.rdbuf(NULL);

  /* Disable I/O synchronization with C library (POSIX I/O is used for
   * stdin/stdout and C++ for stderr)
   */
  std::ios::sync_with_stdio(false);

  /* Initialize OpenSSL and Pango */
  CheckPangoVersion();

  ProgramOptions options;

  /* Parse arguments */
  {
    const int ret = ParseArgs(argc, argv, options);

    if (ret != -1) {
      return ret;
    }
  }

  {
    const std::unique_ptr<const FileDescriptor> input{
        (options.input_filename == kStdioFilename)
            ? std::make_unique<FileDescriptor>(STDIN_FILENO)
            : std::make_unique<PosixFile>(options.input_filename, O_RDONLY,
                                          0400)};

    const std::unique_ptr<const FileDescriptor> output{
        (options.output_filename == kStdioFilename)
            ? std::make_unique<FileDescriptor>(STDOUT_FILENO)
            : std::make_unique<PosixFile>(options.output_filename,
                                          O_WRONLY | O_CREAT | O_TRUNC, 0600)};

    Process(options, input->get(), output->get());
  }

  return EXIT_SUCCESS;
}
}

int main(int argc, char *argv[])
{
  try {
    return QrBackup::Main(argc, argv);

  } catch (const std::exception &exc) {
    std::cerr << "An unhandled error occurred:\n" << exc.what() << std::endl;

    return EXIT_FAILURE;
  }
}

/* vim: set sw=2 sts=2 et : */
