/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <cassert>
#include <algorithm>
#include <array>
#include <stdexcept>
#include <system_error>
#include <iterator>
#include <memory>

#include "qrcode.hpp"

namespace
{
const std::array<std::string, 4> kEcLevelName{"L", "M", "Q", "H"};

struct Deleter
{
  void operator()(QRcode *ptr) const
  {
    ::QRcode_free(ptr);
  }
};
}

namespace QrBackup
{
struct QrCodeImpl
{
  ::QRecLevel eclevel;
  unsigned int version;
  unsigned int width;
  std::vector<bool> bits;
};

QrCode::QrCode(const std::string &data, const ::QRecLevel eclevel,
               const int version)
    : pimpl{std::make_unique<QrCodeImpl>()}
{
  const std::unique_ptr<QRcode, Deleter> qr{
      ::QRcode_encodeString8bit(data.c_str(), version, eclevel)};

  if (!qr) {
    throw std::system_error{errno, std::system_category(),
                            "QRcode_encodeString8bit"};
  }

  pimpl->eclevel = eclevel;
  pimpl->width = qr->width;
  pimpl->version = version;

  /* Build a container of booleans, each describing the state of a module in
   * a QR code (true means black, false means white). The container has
   * (width * width) items.
   */
  std::transform(qr->data, qr->data + (qr->width * qr->width),
                 std::back_inserter(pimpl->bits),
                 [](const char value) { return !!(value & 1); });

  assert(pimpl->bits.size() == (pimpl->width * pimpl->width));
}

QrCode::~QrCode()
{
}

unsigned int QrCode::width() const
{
  return pimpl->width;
}

unsigned int QrCode::version() const
{
  return pimpl->version;
}

std::string QrCode::eclevel() const
{
  bool status{false};
  std::string name;

  std::tie(status, name) = GetEcLevelName(pimpl->eclevel);

  if (!status) {
    throw std::runtime_error("Can't determine error correction level name");
  }

  return name;
}

std::vector<bool> QrCode::data() const
{
  return pimpl->bits;
}

std::vector<std::string> GetEcLevelNames()
{
  return std::vector<std::string>(std::begin(kEcLevelName),
                                  std::end(kEcLevelName));
}

/* Get name of a QR code error correction level.
 */
std::pair<bool, std::string> GetEcLevelName(const ::QRecLevel &eclevel)
{
  if (eclevel >= 0 && eclevel < kEcLevelName.size()) {
    return {true, kEcLevelName[eclevel]};
  }

  return {false, "UNKNOWN"};
}

/* Resolve name of QR code error correction level and get its numeric
 * representation.
 */
std::pair<bool, ::QRecLevel> LookupEcLevel(const std::string &name)
{
  const auto begin = std::begin(kEcLevelName), end = std::end(kEcLevelName);

  const auto it = std::find(begin, end, name);

  if (it == end) {
    return {false, kEcLevelInvalid};
  }

  return {true, static_cast<QRecLevel>(std::distance(begin, it))};
}
}

/* vim: set sw=2 sts=2 et : */
