/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <system_error>
#include <string>
#include <vector>

#include <cstdlib>
#include <cstring>
#include <ctime>
#include <cerrno>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <libgen.h>

#include "utils.hpp"

namespace QrBackup
{
/* Extracts the program name from argv[0].
 */
const std::string GetProgramName(const char *argv0)
{
  std::vector<char> copy(argv0, argv0 + ::strlen(argv0) + 1);

  return ::basename(&copy[0]);
}

/* Formats a timestamp in the ISO 8601 format.
 *
 * Example: 2002-08-14T13:10:20+0200
 */
const std::string FormatTimeWithTz(const time_t ts)
{
  char buffer[64];
  struct tm tm;

  if (!::localtime_r(&ts, &tm)) {
    throw std::system_error{errno, std::system_category(), "localtime_r"};
  }

  size_t ret = ::strftime(buffer, sizeof(buffer), "%FT%T%z", &tm);
  if (ret == 0) {
    throw std::runtime_error("strftime failed");
  }

  return buffer;
}

/* Formats the current time using the ISO 8601 format.
 */
const std::string CurrentTimeWithTz()
{
  return FormatTimeWithTz(::time(NULL));
}

/* Attempts to reserve enough space in a vector if the given POSIX file
 * descriptor points to a regular file whose size can be determined.
 */
void ReserveForFileSize(const int fd, std::vector<unsigned char> &buffer)
{
  struct stat st;

  if (::fstat(fd, &st) == 0 && S_ISREG(st.st_mode)) {
    /* Reserve enough memory */
    buffer.reserve(st.st_size);
  }
}

/* Reads an entire file into a vector.
 */
void ReadFile(const int fd, std::vector<unsigned char> &content)
{
  unsigned char buffer[4096];

  ReserveForFileSize(fd, content);

  while (1) {
    ssize_t length = ::read(fd, buffer, sizeof(buffer));

    if (length == 0) {
      /* EOF */
      break;

    } else if (length < 0) {
      if (errno == EINTR) {
        /* Caught signal */
        continue;
      }

      /* Error */
      throw std::system_error{errno, std::system_category(), "read"};

    } else {
      content.insert(content.end(), buffer, buffer + length);
    }
  }
}
}

/* vim: set sw=2 sts=2 et : */
