/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string>
#include <vector>
#include <array>

#define BOOST_TEST_MODULE Base64
#include <boost/test/unit_test.hpp>

#include "../src/macros.hpp"
#include "../src/base64.hpp"

BOOST_AUTO_TEST_CASE(TestCalcBase64Length)
{
  using ::QrBackup::CalcBase64Length;
  using ::QrBackup::CalcBase64Capacity;

  for (size_t i = 0; i < 1000; ++i) {
    size_t length = CalcBase64Length(i);

    BOOST_CHECK_EQUAL(length % 4, 0);

    BOOST_CHECK_EQUAL(CalcBase64Length(CalcBase64Capacity(length)), length);
  }
}

BOOST_AUTO_TEST_CASE(TestEncodeBase64Empty)
{
  using ::QrBackup::EncodeBase64;

  std::array<unsigned char, 0> data;

  BOOST_CHECK_EQUAL(EncodeBase64(&data[0], 0), "");
}

BOOST_AUTO_TEST_CASE(TestEncodeBase64Trivial)
{
  using ::QrBackup::EncodeBase64;

  {
    const std::array<unsigned char, 1> data{'H'};

    BOOST_CHECK_EQUAL(EncodeBase64(&data[0], data.size()), "SA==");
  }

  {
    const std::array<unsigned char, 2> data{'0', '1'};

    BOOST_CHECK_EQUAL(EncodeBase64(&data[0], data.size()), "MDE=");
  }

  {
    const std::array<unsigned char, 3> data{'A', 'A', 'A'};

    BOOST_CHECK_EQUAL(EncodeBase64(&data[0], data.size()), "QUFB");
  }

  {
    const std::array<unsigned char, 4> data{'X', 'Y', 'Z', '0'};

    BOOST_CHECK_EQUAL(EncodeBase64(&data[0], data.size()), "WFlaMA==");
  }
}

namespace
{
const char kAllByteValuesInBase64[]{
    "AAECAwQFBgcICQoLDA0ODxAREhMUFRYXGBkaGxwdHh8gISIjJCUmJygpKissLS4vMDEyMzQ1"
    "Njc4OTo7PD0+P0BBQkNERUZHSElKS0xNTk9QUVJTVFVWV1hZWltcXV5fYGFiY2RlZmdoaWpr"
    "bG1ub3BxcnN0dXZ3eHl6e3x9fn+AgYKDhIWGh4iJiouMjY6PkJGSk5SVlpeYmZqbnJ2en6Ch"
    "oqOkpaanqKmqq6ytrq+wsbKztLW2t7i5uru8vb6/wMHCw8TFxsfIycrLzM3Oz9DR0tPU1dbX"
    "2Nna29zd3t/g4eLj5OXm5+jp6uvs7e7v8PHy8/T19vf4+fr7/P3+/w=="};
}

BOOST_AUTO_TEST_CASE(TestEncodeBase64AllAsciiChars)
{
  using ::QrBackup::EncodeBase64;

  std::vector<unsigned char> data;

  for (int i = 0; i < 256; ++i) {
    data.emplace_back(i);
  }

  BOOST_CHECK_EQUAL(EncodeBase64(&data[0], data.size()),
                    kAllByteValuesInBase64);
}

/* vim: set sw=2 sts=2 et : */
