/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string>
#include <map>

#define BOOST_TEST_MODULE Hash
#include <boost/test/unit_test.hpp>

#include <openssl/evp.h>

#include "../src/macros.hpp"
#include "../src/digest.hpp"
#include "utils.hpp"

BOOST_AUTO_TEST_CASE(TestCalcDigest)
{
  using QrBackup::HashAlgorithm;
  using QrBackup::CalcDigest;

  const char buffer[]{"Hello World!"};

  const auto result = CalcDigest(buffer, sizeof(buffer) - 1);

  /* At least two algorithms are required */
  BOOST_CHECK_GT(result.size(), 2);

  const std::map<HashAlgorithm, std::string> expected{
#ifndef OPENSSL_NO_MD5
      {NID_md5, "ed076287532e86365e841e92bfc50d8c"},
#endif
#ifndef OPENSSL_NO_SHA
      {NID_sha1, "2ef7bde608ce5404e97d5f042f95f89f1c232871"},
#endif
#ifndef OPENSSL_NO_SHA256
      {NID_sha256,
       "7f83b1657ff1fc53b92dc18148a1d65dfc2d4b1fa3d677284addd200126d9069"},
#endif
#ifndef OPENSSL_NO_WHIRLPOOL
      {NID_whirlpool,
       "d4b3ad3619bc70157376c5426b558dbdad30654cf441ab21d7c08e993873256becc80f3"
       "2448d0218d5b1aab30bf4209e20e3928df002d3cbcfbe501a184680a8"},
#endif
#ifndef OPENSSL_NO_RIPEMD
      {NID_ripemd160, "8476ee4631b9b30ac2754b0ee0c47e161d3f724c"},
#endif
  };

  BOOST_CHECK(result == expected);
}

BOOST_AUTO_TEST_CASE(TestHexifyEmpty)
{
  const unsigned char data[] = {0};

  BOOST_CHECK_EQUAL(QrBackup::Hexify(data, 0), "");
}

BOOST_AUTO_TEST_CASE(TestHexifyFourBytes)
{
  std::array<unsigned char, 4> data{0, 1, 2, 3};

  BOOST_CHECK_EQUAL(QrBackup::Hexify(&data[0], data.size()), "00010203");
}

BOOST_AUTO_TEST_CASE(TestHexifyLargerNumbers)
{
  std::array<unsigned char, 6> data{0x01, 0x94, 0x32, 0x44, 0xff, 0x29};

  BOOST_CHECK_EQUAL(QrBackup::Hexify(&data[0], data.size()), "01943244ff29");
}

/* vim: set sw=2 sts=2 et : */
