/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <algorithm>
#include <functional>
#include <deque>
#include <ostream>

#include <boost/function.hpp>

#define BOOST_TEST_MODULE File
#include <boost/test/unit_test.hpp>

#include "../src/macros.hpp"
#include "../src/file.hpp"

#include "utils.hpp"

namespace
{
/* MD5 digest of buffer containing byte values 0-255 (length 256) */
const std::string kMd5AllByteValues{"e2c865db4162bed963bfaa9ef6ac18f0"};

/* MD5 digest of string "Test" */
const std::string kMd5Test{"0cbc6611f5540bd0809a388dc95a615b"};
}

BOOST_AUTO_TEST_CASE(TestEmptyFile)
{
  using QrBackup::FileContext;

  const FileContext fctx{QR_ECLEVEL_M, "", {}};

  BOOST_CHECK_EQUAL(fctx.length(), 0);
  BOOST_CHECK(fctx.description().empty());
  BOOST_CHECK(!fctx.content_digest().empty());
  BOOST_CHECK(fctx.parts().empty());
}

static std::vector<unsigned char> MakeTestData(const std::string &data)
{
  return std::vector<unsigned char>(data.begin(), data.end());
}

BOOST_AUTO_TEST_CASE(TestShortContent)
{
  using namespace QrBackup;

  std::vector<unsigned char> content(MakeTestData("Test"));

  const FileContext fctx{QR_ECLEVEL_L, "Description", content};

  BOOST_CHECK_EQUAL(fctx.description(), "Description");
  BOOST_CHECK_EQUAL(fctx.length(), 4);
  BOOST_CHECK_GT(fctx.content_digest().size(), 0);

  const auto parts = fctx.parts();

  BOOST_CHECK_EQUAL(parts.size(), 1);

  const auto &part0 = *parts[0];

  BOOST_CHECK_EQUAL(part0.offset, 0);
  BOOST_CHECK_EQUAL(part0.length, 4);
  BOOST_CHECK_EQUAL(part0.qrcode->version(), 1);
  BOOST_CHECK_EQUAL(part0.qrcode->width(), 21);

  /* Top-left corner is always black */
  BOOST_CHECK(part0.qrcode->data()[0]);

  CheckMd5Digest(fctx.content_digest(), kMd5Test);
  CheckMd5Digest(part0.digest, kMd5Test);
}

BOOST_AUTO_TEST_CASE(TestBinaryValues)
{
  using namespace QrBackup;

  std::vector<unsigned char> data;

  for (unsigned int i = 0; i < 256; ++i) {
    data.push_back(i);
  }

  const FileContext fctx{QR_ECLEVEL_L, "Binary data", data};

  BOOST_CHECK_EQUAL(fctx.description(), "Binary data");
  BOOST_CHECK_EQUAL(fctx.length(), 256);
  BOOST_CHECK_GT(fctx.content_digest().size(), 0);

  const auto parts = fctx.parts();

  BOOST_CHECK_EQUAL(parts.size(), 1);

  const FilePart &part0 = *parts[0];

  BOOST_CHECK_EQUAL(part0.qrcode->version(), 12);
  BOOST_CHECK_EQUAL(part0.qrcode->width(), 65);

  CheckMd5Digest(fctx.content_digest(), kMd5AllByteValues);
  CheckMd5Digest(part0.digest, kMd5AllByteValues);
}

namespace
{
const std::array<std::string, 19> kMd5Split{
    "0ec7804b9478253e61cc46f16bfee5b7", /* 2214 bytes */
    "3fc378bbaffbf48954a5e92be980e221", /* 2214 bytes */
    "b65fe9a80aa3a4a7ddbb673629eb4922", /* 2214 bytes */
    "f467b72e1711180e67ef861a494e2f4c", /* 2214 bytes */
    "a5f64d388621c8ace9da6bcf90f3a8e0", /* 2214 bytes */
    "0ec7804b9478253e61cc46f16bfee5b7", /* 2214 bytes */
    "3fc378bbaffbf48954a5e92be980e221", /* 2214 bytes */
    "b65fe9a80aa3a4a7ddbb673629eb4922", /* 2214 bytes */
    "f467b72e1711180e67ef861a494e2f4c", /* 2214 bytes */
    "a5f64d388621c8ace9da6bcf90f3a8e0", /* 2214 bytes */
    "0ec7804b9478253e61cc46f16bfee5b7", /* 2214 bytes */
    "3fc378bbaffbf48954a5e92be980e221", /* 2214 bytes */
    "b65fe9a80aa3a4a7ddbb673629eb4922", /* 2214 bytes */
    "f467b72e1711180e67ef861a494e2f4c", /* 2214 bytes */
    "a5f64d388621c8ace9da6bcf90f3a8e0", /* 2214 bytes */
    "0ec7804b9478253e61cc46f16bfee5b7", /* 2214 bytes */
    "3fc378bbaffbf48954a5e92be980e221", /* 2214 bytes */
    "b65fe9a80aa3a4a7ddbb673629eb4922", /* 2214 bytes */
    "5d3b51f663b94f1c9f590d2edbcbaa2b", /* 1413 bytes */
};
}

BOOST_AUTO_TEST_CASE(TestSplit)
{
  using namespace QrBackup;

  const size_t count = 2751;
  const std::string testtext = "This is a test\n";

  std::vector<unsigned char> data;

  for (unsigned int i = 0; i < count; ++i) {
    std::copy(testtext.begin(), testtext.end(), std::back_inserter(data));
  }

  const FileContext fctx{QR_ECLEVEL_L, "Split data", data};

  BOOST_CHECK_EQUAL(fctx.description(), "Split data");
  BOOST_CHECK_EQUAL(fctx.length(), testtext.length() * count);
  BOOST_CHECK_GT(fctx.content_digest().size(), 0);
  CheckMd5Digest(fctx.content_digest(), "eedbdc9f10f5bd1c32909985dd185f9a");

  const auto parts = fctx.parts();

  BOOST_CHECK_EQUAL(parts.size(), kMd5Split.size());

  /* Check digests of parts */
  auto expected = std::begin(kMd5Split);

  for (const auto &i : parts) {
    CheckMd5Digest(i->digest, *(expected++));
  }
}

/* vim: set sw=2 sts=2 et : */
