/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string>
#include <vector>
#include <sstream>

#define BOOST_TEST_MODULE Utilities
#include <boost/test/unit_test.hpp>
#include <boost/algorithm/string/predicate.hpp>

#include "../src/macros.hpp"
#include "../src/options.hpp"

static std::vector<char> MakeCharVector(const char *text)
{
  return std::vector<char>(text, text + ::strlen(text) + 1);
}

BOOST_AUTO_TEST_CASE(TestUsage)
{
  std::ostringstream buf;

  QrBackup::Usage(buf, "{program}");

  const std::string output = buf.str();

  BOOST_CHECK(boost::starts_with(output, "Usage: {program}"));
  BOOST_CHECK_NE(output.find("(one of L, M, Q, H; defaults to M)"),
                 std::string::npos);
}

BOOST_AUTO_TEST_CASE(TestParseArgsNoOptions)
{
  using namespace QrBackup;

  ProgramOptions options;
  std::ostringstream output;

  char *const dummy[] = {NULL};

  const int result = ParseArgs(0, dummy, options, output, output);

  BOOST_CHECK_EQUAL(result, EXIT_FAILURE);
  BOOST_CHECK(options.description.empty());
  BOOST_CHECK(options.input_filename.empty());
  BOOST_CHECK(options.output_filename.empty());
  BOOST_CHECK_EQUAL(options.eclevel, QR_ECLEVEL_M);
  BOOST_CHECK(boost::starts_with(output.str(), "Usage: (unknown)"));
}

BOOST_AUTO_TEST_CASE(TestParseArgsNoFiles)
{
  using namespace QrBackup;

  ProgramOptions options;
  std::ostringstream output;

  std::vector<char> arg0(MakeCharVector("program"));
  std::vector<char> arg1(MakeCharVector("-d"));
  std::vector<char> arg2(MakeCharVector("Foobar"));

  char *const args[] = {&arg0[0], &arg1[0], &arg2[0]};

  const int result = ParseArgs(ARRAY_SIZE(args), args, options, output, output);

  BOOST_CHECK_EQUAL(result, EXIT_FAILURE);
  BOOST_CHECK(options.input_filename.empty());
  BOOST_CHECK(options.output_filename.empty());
  BOOST_CHECK_EQUAL(options.description, "Foobar");
  BOOST_CHECK_EQUAL(options.eclevel, QR_ECLEVEL_M);
  BOOST_CHECK(boost::starts_with(output.str(), "Usage: program"));
}

BOOST_AUTO_TEST_CASE(TestParseArgsUnknownArgument)
{
  using namespace QrBackup;

  ProgramOptions options;
  std::ostringstream output;

  std::vector<char> arg0(MakeCharVector("program"));
  std::vector<char> arg1(MakeCharVector("-X"));

  char *const args[] = {&arg0[0], &arg1[0]};

  const int result = ParseArgs(ARRAY_SIZE(args), args, options, output, output);

  BOOST_CHECK_EQUAL(result, EXIT_FAILURE);
  BOOST_CHECK(options.description.empty());
  BOOST_CHECK(options.input_filename.empty());
  BOOST_CHECK(options.output_filename.empty());
  BOOST_CHECK_EQUAL(options.eclevel, QR_ECLEVEL_M);
  BOOST_CHECK(boost::starts_with(output.str(), "Usage: program"));
}

BOOST_AUTO_TEST_CASE(TestParseArgsHelp)
{
  using namespace QrBackup;

  ProgramOptions options;
  std::ostringstream output;

  std::vector<char> arg0(MakeCharVector("programforhelp"));
  std::vector<char> arg1(MakeCharVector("-h"));

  char *const args[] = {&arg0[0], &arg1[0]};

  const int result = ParseArgs(ARRAY_SIZE(args), args, options, output, output);

  BOOST_CHECK_EQUAL(result, EXIT_SUCCESS);
  BOOST_CHECK(options.description.empty());
  BOOST_CHECK(options.input_filename.empty());
  BOOST_CHECK(options.output_filename.empty());
  BOOST_CHECK_EQUAL(options.eclevel, QR_ECLEVEL_M);
  BOOST_CHECK(boost::starts_with(output.str(), "Usage: programforhelp"));
}

BOOST_AUTO_TEST_CASE(TestParseArgsVersion)
{
  using namespace QrBackup;

  ProgramOptions options;
  std::ostringstream output;

  std::vector<char> arg0(MakeCharVector("programforversion"));
  std::vector<char> arg1(MakeCharVector("-V"));

  char *const args[] = {&arg0[0], &arg1[0]};

  const int result = ParseArgs(ARRAY_SIZE(args), args, options, output, output);

  BOOST_CHECK_EQUAL(result, EXIT_SUCCESS);
  BOOST_CHECK(options.description.empty());
  BOOST_CHECK(options.input_filename.empty());
  BOOST_CHECK(options.output_filename.empty());
  BOOST_CHECK_EQUAL(options.eclevel, QR_ECLEVEL_M);
  BOOST_CHECK_EQUAL(output.str(), "programforversion " PACKAGE_VERSION "\n");
}

BOOST_AUTO_TEST_CASE(TestParseArgsWrongEcLevel)
{
  using namespace QrBackup;

  ProgramOptions options;
  std::ostringstream output;

  std::vector<char> arg0(MakeCharVector("wronglevel"));
  std::vector<char> arg1(MakeCharVector("-l"));
  std::vector<char> arg2(MakeCharVector("X"));

  char *const args[] = {&arg0[0], &arg1[0], &arg2[0]};

  const int result = ParseArgs(ARRAY_SIZE(args), args, options, output, output);

  BOOST_CHECK_EQUAL(result, EXIT_FAILURE);
  BOOST_CHECK(options.description.empty());
  BOOST_CHECK(options.input_filename.empty());
  BOOST_CHECK(options.output_filename.empty());
  BOOST_CHECK_EQUAL(options.eclevel, QR_ECLEVEL_M);
  BOOST_CHECK(boost::starts_with(output.str(), "Usage: wronglevel"));
}

BOOST_AUTO_TEST_CASE(TestParseArgsWithFiles)
{
  using namespace QrBackup;

  std::vector<char> arg0(MakeCharVector("programwithfiles"));
  std::vector<char> arg1(MakeCharVector("-l"));
  std::vector<char> arg2(MakeCharVector("H"));
  std::vector<char> arg3(MakeCharVector("path1"));
  std::vector<char> arg4(MakeCharVector("path2"));

  char *const args[] = {&arg0[0], &arg1[0], &arg2[0], &arg3[0], &arg4[0]};

  ProgramOptions options;
  std::ostringstream output;

  const int result = ParseArgs(ARRAY_SIZE(args), args, options, output, output);

  BOOST_CHECK_EQUAL(result, -1);
  BOOST_CHECK(options.description.empty());
  BOOST_CHECK_EQUAL(options.eclevel, QR_ECLEVEL_H);
  BOOST_CHECK_EQUAL(options.input_filename, "path1");
  BOOST_CHECK_EQUAL(options.output_filename, "path2");
  BOOST_CHECK(output.str().empty());
}

BOOST_AUTO_TEST_CASE(TestParseArgsStdin)
{
  using namespace QrBackup;

  std::vector<char> arg0(MakeCharVector("progstdin"));
  std::vector<char> arg1(MakeCharVector("-"));
  std::vector<char> arg2(MakeCharVector("output"));

  char *const args[] = {&arg0[0], &arg1[0], &arg2[0]};

  ProgramOptions options;
  std::ostringstream output;

  const int result = ParseArgs(ARRAY_SIZE(args), args, options, output, output);

  BOOST_CHECK_EQUAL(result, -1);
  BOOST_CHECK(options.description.empty());
  BOOST_CHECK_EQUAL(options.input_filename, "-");
  BOOST_CHECK_EQUAL(options.output_filename, "output");
  BOOST_CHECK_EQUAL(options.eclevel, QR_ECLEVEL_M);
  BOOST_CHECK(output.str().empty());
}

BOOST_AUTO_TEST_CASE(TestParseArgsTooMany)
{
  using namespace QrBackup;

  std::vector<char> arg0(MakeCharVector("toomanyargs"));
  std::vector<char> arg1(MakeCharVector("arg1"));
  std::vector<char> arg2(MakeCharVector("arg2"));
  std::vector<char> arg3(MakeCharVector("arg3"));

  char *const args[] = {&arg0[0], &arg1[0], &arg2[0], &arg3[0]};

  ProgramOptions options;
  std::ostringstream output;

  const int result = ParseArgs(ARRAY_SIZE(args), args, options, output, output);

  BOOST_CHECK_EQUAL(result, EXIT_FAILURE);
  BOOST_CHECK(options.description.empty());
  BOOST_CHECK(options.input_filename.empty());
  BOOST_CHECK(options.output_filename.empty());
  BOOST_CHECK_EQUAL(options.eclevel, QR_ECLEVEL_M);
  BOOST_CHECK(boost::starts_with(output.str(), "Usage: toomanyargs"));
}

BOOST_AUTO_TEST_CASE(TestParseArgsStdinNoOutput)
{
  using namespace QrBackup;

  std::vector<char> arg0(MakeCharVector("stdinnooutput"));
  std::vector<char> arg1(MakeCharVector("-"));

  char *const args[] = {&arg0[0], &arg1[0]};

  ProgramOptions options;
  std::ostringstream output;

  const int result = ParseArgs(ARRAY_SIZE(args), args, options, output, output);

  BOOST_CHECK_EQUAL(result, EXIT_FAILURE);
  BOOST_CHECK(options.description.empty());
  BOOST_CHECK(options.input_filename.empty());
  BOOST_CHECK(options.output_filename.empty());
  BOOST_CHECK_EQUAL(options.eclevel, QR_ECLEVEL_M);
  BOOST_CHECK(boost::starts_with(output.str(), "Usage: stdinnooutput"));
}

/* vim: set sw=2 sts=2 et : */
