/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <system_error>

#define BOOST_TEST_MODULE QrCode
#include <boost/test/unit_test.hpp>

#include <qrencode.h>

#include "../src/qrcode.hpp"

BOOST_AUTO_TEST_CASE(TestQrCodeWithoutData)
{
  using QrBackup::QrCode;

  for (int version = 1; version < QRSPEC_VERSION_MAX; ++version) {
    BOOST_CHECK_THROW(QrCode("", QR_ECLEVEL_L, version), std::system_error);
  }
}

BOOST_AUTO_TEST_CASE(TestQrCode)
{
  using QrBackup::QrCode;

  const std::string buffer{"Hello World"};

  for (int version = 1; version < QRSPEC_VERSION_MAX; ++version) {
    const QrCode qr{buffer, QR_ECLEVEL_M, version};
    BOOST_CHECK_EQUAL(qr.version(), version);
    BOOST_CHECK_EQUAL(qr.width(), 17 + (4 * version));

    /* Top-left corner is always black */
    BOOST_CHECK(qr.data()[0]);
  }
}

BOOST_AUTO_TEST_CASE(TestLookupEcLevel)
{
  using QrBackup::GetEcLevelName;
  using QrBackup::kEcLevelInvalid;
  using QrBackup::LookupEcLevel;

  for (const auto &i : {"", "longer", "l", "m"}) {
    BOOST_CHECK(LookupEcLevel(i) == std::make_pair(false, kEcLevelInvalid));
  }

  const std::pair<std::string, QRecLevel> map[]{
      {"M", QR_ECLEVEL_M},
      {"H", QR_ECLEVEL_H},
      {"Q", QR_ECLEVEL_Q},
      {"L", QR_ECLEVEL_L},
  };

  for (const auto &i : map) {
    BOOST_CHECK(LookupEcLevel(i.first) == std::make_pair(true, i.second));
    BOOST_CHECK(GetEcLevelName(i.second) == std::make_pair(true, i.first));
  }
}

BOOST_AUTO_TEST_CASE(TestGetEcLevelNames)
{
  using QrBackup::GetEcLevelNames;

  const std::string expected[]{"L", "M", "Q", "H"};

  const auto names = GetEcLevelNames();

  BOOST_CHECK_EQUAL_COLLECTIONS(std::begin(names), std::end(names),
                                std::begin(expected), std::end(expected));
}

BOOST_AUTO_TEST_CASE(TestGetEcLevelName)
{
  using QrBackup::GetEcLevelName;
  using QrBackup::kEcLevelInvalid;

  const QRecLevel values[]{
      kEcLevelInvalid, static_cast<QRecLevel>(QR_ECLEVEL_H + 1),
  };

  for (const auto i : values) {
    BOOST_CHECK(GetEcLevelName(i) ==
                std::make_pair(false, std::string{"UNKNOWN"}));
  }
}

/* vim: set sw=2 sts=2 et : */
