/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string>
#include <sstream>
#include <system_error>
#include <vector>
#include <memory>

#include <cstdlib>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

#define BOOST_TEST_MODULE Utilities
#include <boost/test/unit_test.hpp>
#include <boost/filesystem.hpp>

#include "../src/macros.hpp"
#include "../src/utils.hpp"
#include "../src/posixfile.hpp"
#include "utils.hpp"

class WithEnvironment
{
public:
  WithEnvironment() = delete;
  WithEnvironment(const WithEnvironment &) = delete;

  WithEnvironment(const std::string &name, const std::string &value)
      : _name{name}, _has_original{false}
  {
    /* Attempt to retrieve original value */
    {
      const char *orig_value = ::getenv(name.c_str());

      if (orig_value) {
        _has_original = true;
        _original.assign(orig_value);
      }
    }

    ::setenv(name.c_str(), value.c_str(), 1);
  }

  virtual ~WithEnvironment()
  {
    if (_has_original) {
      /* Restore original value */
      ::setenv(_name.c_str(), _original.c_str(), 1);
    } else {
      /* Variable was not set */
      ::unsetenv(_name.c_str());
    }
  }

  WithEnvironment &operator=(const WithEnvironment &) = delete;

private:
  const std::string _name;

  bool _has_original;
  std::string _original;
};

class WithTimezone final : public WithEnvironment
{
public:
  WithTimezone() = delete;
  WithTimezone(const WithTimezone &) = delete;

  WithTimezone(const std::string &tz) : WithEnvironment{"TZ", tz}
  {
    ::tzset();
  }

  WithTimezone &operator=(const WithTimezone &) = delete;
};

BOOST_AUTO_TEST_CASE(FormatEpochUTC)
{
  const WithTimezone tz{"UTC"};

  BOOST_CHECK_EQUAL(QrBackup::FormatTimeWithTz(0), "1970-01-01T00:00:00+0000");
  BOOST_CHECK_EQUAL(QrBackup::FormatTimeWithTz(1212140759),
                    "2008-05-30T09:45:59+0000");
  BOOST_CHECK_EQUAL(QrBackup::FormatTimeWithTz(1364073159),
                    "2013-03-23T21:12:39+0000");
}

BOOST_AUTO_TEST_CASE(FormatEpochEuropeBerlin)
{
  const WithTimezone tz{"Europe/Berlin"};

  BOOST_CHECK_EQUAL(QrBackup::FormatTimeWithTz(0), "1970-01-01T01:00:00+0100");
  BOOST_CHECK_EQUAL(QrBackup::FormatTimeWithTz(140623692),
                    "1974-06-16T15:08:12+0100");
}

BOOST_AUTO_TEST_CASE(FormatEpochAmericaLosAngeles)
{
  const WithTimezone tz{"America/Los_Angeles"};

  BOOST_CHECK_EQUAL(QrBackup::FormatTimeWithTz(0), "1969-12-31T16:00:00-0800");
  BOOST_CHECK_EQUAL(QrBackup::FormatTimeWithTz(124432249),
                    "1973-12-10T20:30:49-0800");
  BOOST_CHECK_EQUAL(QrBackup::FormatTimeWithTz(1364073598),
                    "2013-03-23T14:19:58-0700");
}

BOOST_AUTO_TEST_CASE(TestCurrentTimeWithTz)
{
  BOOST_CHECK(!::QrBackup::CurrentTimeWithTz().empty());
}

BOOST_AUTO_TEST_CASE(TestGetProgramName)
{
  using ::QrBackup::GetProgramName;

  BOOST_CHECK_EQUAL(GetProgramName("foo"), "foo");
  BOOST_CHECK_EQUAL(GetProgramName("/tmp/bar"), "bar");
}

BOOST_AUTO_TEST_CASE(TestReadFileEmpty)
{
  using ::QrBackup::ReadFile;
  using ::QrBackup::TemporaryFile;

  TemporaryFile tmpfile;

  std::vector<unsigned char> content;

  ReadFile(tmpfile.get(), content);

  BOOST_CHECK(content.empty());
}

BOOST_AUTO_TEST_CASE(TestReadFileWithContent)
{
  using ::QrBackup::ReadFile;
  using ::QrBackup::ReserveForFileSize;
  using ::QrBackup::TemporaryFile;

  TemporaryFile tmpfile;

  const unsigned int count = 5000;
  const std::string testdata{"Hello World 1234\n"};

  for (unsigned int i = 0; i < count; ++i) {
    if (::write(tmpfile.get(), testdata.c_str(), testdata.size()) < 0) {
      BOOST_ERROR("Failed to write test file");
    }
  }

  /* Check if memory is reserved */
  {
    const off_t curpos = ::lseek(tmpfile.get(), 0, SEEK_CUR);

    BOOST_REQUIRE_EQUAL(curpos, count * testdata.size());

    std::vector<unsigned char> tmp;

    BOOST_CHECK_LT(tmp.capacity(), curpos);

    ReserveForFileSize(tmpfile.get(), tmp);

    BOOST_CHECK_GE(tmp.capacity(), curpos);
  }

  ::lseek(tmpfile.get(), 0, SEEK_SET);

  /* Read whole temporary file */
  std::vector<unsigned char> content;

  ReadFile(tmpfile.get(), content);

  {
    std::vector<unsigned char> expected;

    /* Build expected value */
    for (unsigned int i = 0; i < count; ++i) {
      std::copy(testdata.begin(), testdata.end(), std::back_inserter(expected));
    }

    BOOST_CHECK_EQUAL_COLLECTIONS(expected.begin(), expected.end(),
                                  content.begin(), content.end());
  }
}

BOOST_AUTO_TEST_CASE(TestReadFileError)
{
  using ::QrBackup::ReadFile;
  using ::QrBackup::TemporaryFile;

  std::vector<unsigned char> content;

  /* Use bad file descriptor deliberately */
  const int fd = TemporaryFile().get();

  BOOST_CHECK_THROW(ReadFile(fd, content), std::system_error);
}

BOOST_AUTO_TEST_CASE(TestWrapMkstemp)
{
  using ::QrBackup::WrapMkstemp;

  BOOST_CHECK_THROW(WrapMkstemp({}), std::runtime_error);

  {
    WithTemporaryDirectory tmpdir;

    const std::vector<std::string> templates{
        /* Iinvalid template in a non-existing directory */
        (tmpdir.get() / "foo" / "bar").native(),
    };

    BOOST_CHECK_THROW(WrapMkstemp(templates), std::system_error);
  }
}

/* vim: set sw=2 sts=2 et : */
