/*
 * Copyright (c) 2013, Michael Hanselmann.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - Neither the name of the copyright holder nor the names of contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <system_error>
#include <stdexcept>
#include <cstdlib>

#undef BOOST_TEST_MAIN
#include <boost/test/unit_test.hpp>

#include <openssl/evp.h>

#include "../src/macros.hpp"
#include "utils.hpp"

void CheckMd5Digest(
    const std::map<QrBackup::HashAlgorithm, std::string> &digest,
    const std::string &expected)
{
  using QrBackup::HashAlgorithm;

  const HashAlgorithm md5{EVP_MD_nid(EVP_md5())};

  const auto it = digest.find(md5);

  if (it == digest.end()) {
    BOOST_ERROR("Missing MD5 digest");
    return;
  }

  BOOST_CHECK_EQUAL(it->second, expected);
}

std::string WrapMkdtemp(const std::vector<std::string> &templates)
{
  assert(!templates.empty());

  for (auto it = templates.begin();; ++it) {
    std::vector<char> path{std::begin(*it), std::end(*it)};
    path.emplace_back('\0');

    const char *result = ::mkdtemp(&path[0]);

    if (result != NULL) {
      return result;
    }

    if (std::distance(it, templates.end()) == 1) {
      /* Last template */
      throw std::system_error{errno, std::system_category(), "mkdtemp"};
    }
  }

  throw std::logic_error{"Should never reach this line"};
}

WithTemporaryDirectory::WithTemporaryDirectory()
    : _path{WrapMkdtemp(QrBackup::GetTempNameTemplates())}
{
}

WithTemporaryDirectory::~WithTemporaryDirectory()
{
  if (!_path.empty()) {
    boost::filesystem::remove_all(_path);
  }
}

/* vim: set sw=2 sts=2 et : */
